package 
{
    import flash.events.Event;
    import flash.events.IEventDispatcher;
    
    import mx.binding.BindabilityInfo;
    import mx.binding.utils.ChangeWatcher;
    import mx.core.mx_internal;
    import mx.utils.DescribeTypeCache;
    import mx.utils.ObjectProxy;
    
    import org.spicefactory.lib.reflect.ClassInfo;
    import org.spicefactory.parsley.core.registry.ObjectProcessor;
    import org.spicefactory.parsley.core.registry.ObjectProcessorFactory;
    import org.spicefactory.parsley.processor.util.ObjectProcessorFactories;
    
    use namespace mx_internal;
    
    /**
     *  The BindingDependencyProcessor helps process the bindingDependency 
     *  property of the Bindable metadata.
     */
    public class BindingDependencyProcessor implements ObjectProcessor
    {
        //--------------------------------------------------------------------------
        //
        //  Class methods
        //
        //--------------------------------------------------------------------------
        
        /**
         *  Standard parsley Factory signature
         * 
         *  @param property The getter that is becoming bindable
         *  @param dependentProperty The name of the dependent property
         *  @param event The event to use for this binding
         */
        public static function newFactory (property:String, 
                                           dependentProperty:String, 
                                           event:String) : ObjectProcessorFactory 
        {
            var params:Array = [property, dependentProperty, event];
            return ObjectProcessorFactories.newFactory(BindingDependencyProcessor, params);
        }
        
        /**
         *  Add a binding dependency from one property or method onto another 
         *  property or method.
         * 
         *  @param target The instance of the object we're operating on
         *  @param property The getter or method that is becoming bindable
         *  @param dependentProperty The name of the dependent property
         *  @param event The event to use for this binding
         *  @eventDispatcher The event dispatcher to fire when the dependent property changes
         * 
         *  @return the ChangeWatcher used to watch the dependent property
         */
        public static function addBindingDependency (target:IEventDispatcher,
                                                     property:String, 
                                                     dependentProperty:String, 
                                                     event:String = null,
                                                     eventDispatcher:Function = null) : ChangeWatcher 
        {
            // if no event, use property + "Changed"
            if (!event)
                event = property + "Changed";
            
            // see if this event is in the bindable metadata cache
            var info:BindabilityInfo = DescribeTypeCache.describeType(target).bindabilityInfo;
            var events:Object = info.getChangeEvents(property);
            
            // if it's not in the bindable metadata cache, put it there
            if (events[event] === undefined)
                events[event] = true;
            
            if (eventDispatcher == null)
            {
                eventDispatcher = function(eventObj:Event):void
                {
                    target.dispatchEvent(new Event(event.toString()));
                }
            }
            
            // watch for changes on dependentProperty.  When a change happens, dispatch a change event
            return ChangeWatcher.watch(target, dependentProperty.split("."), eventDispatcher);
        }
        
        //--------------------------------------------------------------------------
        //
        //  Constructor
        //
        //--------------------------------------------------------------------------
        
        /**
         *  Constructor
         * 
         *  @param target The instance of the object we're operating on
         *  @param property The getter or method that is becoming bindable
         *  @param dependentProperty The name of the dependent property
         *  @param event The event to use for this binding
         */
        public function BindingDependencyProcessor(target:Object, 
                                                   property:String, 
                                                   dependentProperty:String, 
                                                   event:String)
        {
            super();
            
            this.target = IEventDispatcher(target.instance);
            this.dependentProperty = dependentProperty;
            this.property = property;
            this.event = event;
        }
        
        //--------------------------------------------------------------------------
        //
        //  Private Variables
        //
        //--------------------------------------------------------------------------
        
        /**
         *  @private
         *  The instance of the obejct we're operating on
         */
        private var target:IEventDispatcher;
        
        /**
         *  @private
         *  The getter that is becoming bindable
         */
        private var property:String;
        
        /**
         *  @private
         *  The name of the dependent property in dot-notation
         */
        private var dependentProperty:String;
        
        /**
         *  @private
         *  The event to use for this binding
         */
        private var event:String;
        
        /**
         *  @private
         *  The instance of the ChangeWatcher used to watch for 
         *  dependent property changes.  We keep a reference to this so 
         *  on destroy, we can stop watching.
         */
        private var dependentPropertyChangeWatcher:ChangeWatcher;
        
        //--------------------------------------------------------------------------
        //
        //  Methods: Object Processor
        //
        //--------------------------------------------------------------------------
        
        /**
         *  @private
         */
        public function preInit () : void
        {
            dependentPropertyChangeWatcher = BindingDependencyProcessor.addBindingDependency(target, 
                property, dependentProperty, event, bindingEventDispatcher);
        }
        
        /**
         *  @private
         */
        public function postDestroy () : void 
        {
            // stop watching the dependentProperty
            if (dependentPropertyChangeWatcher)
                dependentPropertyChangeWatcher.unwatch();
        }
        
        //--------------------------------------------------------------------------
        //
        //  Event Listeners
        //
        //--------------------------------------------------------------------------
        
        /**
         *  @private
         *  Helper method that will run when the 
         *  dependentProperty changes (thanks to the ChangeWatcher).
         * 
         *  <p>This method just dispatches the appropriate binding event.</p>
         */
        private function bindingEventDispatcher(event:Event):void
        {
            target.dispatchEvent(new Event(this.event));
        }
    }
}